-- All code written and maintained by Yewchi 
-- zyewchi@gmail.com

--~ Conditional checking; GameState updates and retreival; Manages SavedVaraibles

-- TODO The tables in this file are intended to be local, but this goes against simplistic segment conditionals, because we require global get functions stated in them to access locals.
-- TODO Because we can't require files, all functions and tables accessed from the function strings of segment lengths or segment conditionals need to be local to the GameStateInterface. Ideally, segments will eventually contain their own language, with keywords such that statements like "thisSpell.charges UP" and "auraPet.timeRemaining < 5" are possible. Then we can move evaulation code to an EvaluationLibrary--listing the functions available to conditionals--which would perform the rules interpretation.
-- TODO Spells data could be generated automatically, however the segement data would need a way to indicate which unit the spell will be cast from/cast on. And a lot of spellIds in spellbooks do not indicate the same spell as a buff applied.

--- Flags --
--
SOUND_PATH = "path"
SOUND_FUNC = "func"
SOUND_PATH_PREFIX = "path::"
SOUND_FUNC_PREFIX = "func::"
--
-- /Flags --

spells = nil
events = nil
spellsSnapshot = nil

GS = {} -- A table for custom segment stored values

-------- SaveSpellSnapshot()
local function SaveSpellSnapshot(spellId)
	if spellsSnapshot[spellId] == nil then spellsSnapshot[spellId] = {} end
	
	for n=1,#spells[spellId],1 do
		spellsSnapshot[spellId][n] = spells[spellId][n]
	end
end

-- TODO FULLY OVERWRITES SavedVariables for spec
-------------- GSI_InitializeAndLoadExample()
local function GSI_InitializeAndLoadExample(specId, funcsForLoading, fullReset)
	if funcsForLoading == nil then return false end
	
	if DEBUG or fullReset or SV_Specializations[specId] == nil then 
		SV_Specializations[specId] = {} 
		SV_Specializations[specId].spells = {}
		SV_Specializations[specId].events = {}
		SV_Specializations[specId].segments = {}
	end
	
	for spellId,array in pairs(funcsForLoading["GetSpells"]()) do
		if SV_Specializations[specId].spells[spellId] == nil then
			SV_Specializations[specId].spells[spellId] = {}
		end
		SV_Specializations[specId].spells[spellId] = array
	end
	for eventId,array in pairs(funcsForLoading["GetEvents"]()) do
		if SV_Specializations[specId].events[eventId] == nil then
			SV_Specializations[specId].events[eventId] = {}
		end
		SV_Specializations[specId].events[eventId] = array
	end
	for segmentKey,array in pairs(funcsForLoading["GetSegments"]()) do
		if SV_Specializations[specId].segments[segmentKey] == nil then
			SV_Specializations[specId].segments[segmentKey] = {}
		end
		SV_Specializations[specId].segments[segmentKey] = array
	end
end

-------- GSI_EvaluateLength()
function GSI_EvaluateLength(length)
	local t = type(length)
	if t == "number" then
		return length
	elseif t == "string" then
		local eval = loadstring(length)()
		if type(eval) == "number" then
			return eval
		else
			return 0.0
		end
	elseif t == "nil" then
		return 0.0
	else
		error({code=ERR_INVALID_CONDITIONAL_RESULT, func="GSI_EvaluateLength(length = "..(length == nil and "nil" or length).." type:"..type(length)..")"})
	end
end

-------- GSI_EvaluateSound()
function GSI_EvaluateSound(sound) -- TODO Memoize soundPaths[(sounds_root_cut)"folder/folder/.../filename"(extension cut)] = "full/file/path.ogg"
	if type(sound) == "number" then
		return sound
	elseif type(sound) == "string" then 
		local split = SplitString(sound, "::")
		if #split == 1 then
			return sound
		elseif #split == 2 then
			if split[1] == SOUND_PATH and split[2] ~= nil then
				return split[2]
			elseif split[1] == SOUND_FUNC and split[2] ~= nil then
				return loadstring(split[2])()
			end
		end
	end
	return nil
end

-------- GSI_EvaluateConditional()
function GSI_EvaluateConditional(conditional)
	if type(conditional) == "string" then 
		return loadstring(conditional)()
	end
	return conditional == true
end

-------- GSI_UpdateSpellTable()
function GSI_UpdateSpellTable(spellId, cdDur, cdExpiration)
	if spells[spellId] == nil then
		error({code=ERR_UNKNOWN_SPELL_AS_ARGUMENT, func="GSI_UpdateSpellTable() spellId="..(spellId~=nil and spellId or "nil").." cdDur="..(cdDur~=nil and cdDur or "nil").." cdExpiration="..(cdExpiration~=nil and cdExpiration or "nil")})
	end
	local thisSpell = spells[spellId]
	
	if cdDur == nil or cdExpiration == nil then
if DEBUG then if cdDur~=cdExpiration then print(audioQsSpecifier..debugSpecifier.."GSI_UpdateSpellTable(): - "..(cdDur == nil and "cdDur" or "cdExpiration").." was sole nil passed.") end end
		local start, dur = GetSpellCooldown(spellId)
		cdDur = dur
		cdExpiration = start + dur
	end
	
	if (thisSpell[SPELL_SPELL_TYPE] == SPELL_TYPE_AURA and thisSpell[SPELL_EXPIRATION] ~= cdExpiration) or
			(thisSpell[SPELL_SPELL_TYPE] == SPELL_TYPE_ABILITY and (thisSpell[SPELL_CHARGES] ~= GetSpellCharges(spellId) or thisSpell[SPELL_EXPIRATION] ~= cdExpiration)) then
		SaveSpellSnapshot(spellId)
		thisSpell[SPELL_CHARGES] = GetSpellCharges(spellId)
		thisSpell[SPELL_DURATION] = cdDur
		thisSpell[SPELL_EXPIRATION] = cdExpiration
		
		return true
	end
	return false
end

-------- GSI_UpdateAllSpellTables()
function GSI_UpdateAllSpellTables()
	for spellId,_ in pairs(spells) do
		local cdStart, cdDur = GetSpellCooldown(spellId)
		local cdExpiration = cdStart + cdDur
		
		GSI_UpdateSpellTable(spellId, cdDur, cdExpiration)
	end
end

-------- GSI_UpdateEventTable(event)
function GSI_UpdateEventTable(event, ...)
	if events[event] == nil then
		error({code=ERR_UNKNOWN_EVENT_AS_ARGUMENT, func="GSI_UpdateEventTable(event="..(event~=nil and event or "nil")..")"})
	end
	
	local args = {...}
	if ... ~= nil then
		for n=1,#args,1 do
			events[event][n] = args[n]
		end
	end
end

-------- GSI_GetSpell()
function GSI_GetSpell(spellId)
	return spells[spellId]
end

-------- GSI_GetSpellsTable()
function GSI_GetSpellsTable()
	return spells
end

-------- GSI_GetSpellsSnapshotTable()
function GSI_GetSpellsSnapshotTable()
	return spellsSnapshot
end

-------- GetAuraInfo() -- I don't want stuff like this under the "GSI_" prefix, because it does not pertain to performing operations on the AudioQ GameState data. It only performs the operation of retreival for a particular peice of information straight from the blizAPI. Having it in the GSI file is only a consequency of not having a more intricate function string parser.
function GetAuraInfo(unitId, spellId, spellType)
	for n=1,40,1 do
		local auraSpellId = select(UNIT_AURA_SPELL_ID, UnitAura(unitId, n, spellType))
		if auraSpellId == nil then 
			return nil
		elseif auraSpellId == spellId then
			return UnitAura(unitId, n, spellType)
		end
	end
end

-------- GSI_GetSegmentsTable()
function GSI_GetSegmentsTable(specIdToLoad)
	if SV_Specializations[specIdToLoad] == nil then
		return nil
	end
	return SV_Specializations[specIdToLoad].segments
end

-------- GSI_AuraIsIncluded()
function GSI_AuraIsIncluded(spellId)
	return false ~= (spells ~= nil and spells[spellId] and spells[spellId][SPELL_SPELL_TYPE] == SPELL_TYPE_AURA or false)
end

-------- GSI_SpecHasPrompts()
function GSI_SpecHasPrompts(specIdToCheck)
	return false ~= (SV_Specializations ~= nil and SV_Specializations[specIdToCheck] or false)
end

-------- GSI_RegisterCustomEvents()
function GSI_RegisterCustomEvents(frame)
if VERBOSE then print(audioQsSpecifier..debugSpecifier.."Registering custom events:") end
	for event,_ in pairs(events) do
		frame:RegisterEvent(event)
if VERBOSE then print(audioQsSpecifier..debugSpecifier.."- "..event) end
	end
end

-------- GSI_UnregisterCustomEvents()
function GSI_UnregisterCustomEvents(frame)
	frame:UnregisterAllEvents(event)
end

-------- GSI_EventIsIncluded()
function GSI_EventIsIncluded(eventToCheck)
	return events[eventToCheck] ~= nil
end

-------- GSI_LoadSpecTables(specId)
function GSI_LoadSpecTables(specId, funcsForLoading)
	spellsSnapshot = {}
	
	if SV_Specializations == nil then
		SV_Specializations = {}
	end
	if type(funcsForLoading) == "table" then
		GSI_InitializeAndLoadExample(specId, funcsForLoading)
	end
	
	if SV_Specializations[specId] ~= nil then 
		spells = SV_Specializations[specId].spells
		events = SV_Specializations[specId].events
		return true
	end
	return false
end